from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import functools
import json
import sys
import base64
import requests
import requests.auth
from .compat import *

class ParseException(Exception):
    """An exception that occurrs while parsing a test specification"""

    def __init__(self, data_type, data_name, message):
        """
        :arg string data_type: The type of item that was being parsed when the
            error occurred - 'test' or 'testcase'
        :arg string data_name: The name of the test or testcase that was being
            parsed when the error occurred
        :arg string message: The error message
        """
        super(ParseException, self).__init__("%s %s: %s" % (data_type, data_name, message))

def _get_field(data_type, data, field_name, parser, required=False):
    """
    Gets/parses a field from a test/testcase
    :arg string data_type: The type of item that is being parsed - 'test' or
        'testcase'
    :arg dict data: The item that contains the field
    :arg string field_name: The name of the field to extract
    :arg function parser: The function used to parse the raw value
    """

    if field_name in data:
        # Get the field value if it exists
        value = data[field_name]
        
        if parser:
            # Parse the field value or throw an error
            try:
                value = parser(value)
            except Exception as e:
                data_name = data.get("name", "unknown")
                raise ParseException(data_type, data_name, "Could not parse field %s: %s" % (field_name, str(e)))

        return value
    elif required:
        # Throw an error if the field is required and does not exist
        data_name = data.get("name", "unknown")
        raise ParseException(data_type, data_name, "Missing required field %s" % field_name)
    else:
        # Return nothing if the field is not required and does not exist
        return None

def _auth_config_parser(config):
    """
    Parses an auth configuration. Two forms are allowed:
    * 'basic user pass' - Performs HTTP basic authentication with the
      specified username and password
    * 'digest user pass' - Performs HTTP digest authentication with the
      specified username and password

    :arg string config: The auth config string
    """
    parts = config.split()

    if len(parts) != 3:
        raise Exception(data_type, data_name, "Invalid auth config. Must specify an auth method (basic or digest) followed by the auth parameters for that method.")

    if parts[0] == "basic":
        return requests.auth.HTTPBasicAuth(parts[1], parts[2])
    elif parts[0] == "digest":
        return requests.auth.HTTPDigestAuth(parts[1], parts[2])
    else:
        raise Exception("Unknown auth method: %s" % parts[0])

def _get_file_contents(path):
    """
    Gets the contents of a specified file, ensuring that the file is properly
    closed when the function exits
    """
    with open(path, "r") as f:
        return f.read()

class TestcaseResult(object):
    """
    Model for the result of a testcase execution. This can be used in a `with`
    block so that the model catches exceptions and temporarily replaces
    stdout/stderr with a string buffer for output capture.
    """

    def __enter__(self):
        # Temporarily replace stdout/stderr with a string buffer
        self._old_stdout = sys.stdout
        self._old_stderr = sys.stderr
        sys.stdout = self._captured_stdout = StringIO()
        sys.stderr = self._captured_stderr = StringIO()
        return self

    def __exit__(self, type, value, traceback):
        # Save the error
        self.error_type = type
        self.error = value
        self.error_traceback = traceback

        # Capture the stdout/stderr results
        self.stdout = self._captured_stdout.getvalue()
        self._captured_stdout.close()

        self.stderr = self._captured_stderr.getvalue()
        self._captured_stderr.close()

        # Set stdout/stderr back to their old values
        sys.stdout = self._old_stdout
        sys.stderr = self._old_stderr

        return True

    @property
    def failed(self):
        """Returns whether the testcase failed"""
        return self.error != None

class Testcase(object):
    """Model that represents a testcase specification"""

    def __init__(self, name):
        """
        Creates a new testcase
        :arg string name: The name of the testcase
        """
        self.name = name

    @classmethod
    def _choose_field(cls, testcase, data, field_type_name, **fields):
        all_fields = ((field, _get_field("testcase", data, field, parser)) for (field, parser) in fields.items())
        set_fields = [(field, value) for (field, value) in all_fields if value != None]

        if len(set_fields) > 1:
            raise ParseException("testcase", testcase.name, "More than one %s defined" % field_type_name)
        else:
            setattr(testcase, "%s_type" % field_type_name, set_fields[0][0] if set_fields else None)
            setattr(testcase, field_type_name, set_fields[0][1] if set_fields else None)

    @classmethod
    def parse(cls, data):
        """
        Parses a testcase into a model
        :arg dict data: The data to parse into a model
        """

        # Create a shortcut for extracting a field value
        field = functools.partial(_get_field, "testcase", data)

        # Get the request fields
        t = cls(field("name", str, required=True))
        t.method = field("method", lambda m: str(m).upper()) or "GET"
        t.url = field("url", str, required=True)
        t.query_params = field("query_params", dict) or {}
        t.headers = field("headers", dict) or {}
        t.auth = field("auth", _auth_config_parser)

        # Get the request body payload
        cls._choose_field(t, data, "body",
            body=lambda b: b,
            form_body=lambda b: urllib.urlencode(dict(b)),
            base64_body=lambda b: base64.b64decode(bytes(b)),
            file_body=_get_file_contents
        )

        # Set the response fields
        t.code = field("code", int)
        t.response_url = field("response_url", str)
        t.response_headers = field("response_headers", dict) or {}

        # Set the expected response body
        cls._choose_field(t, data, "response_body",
            response_body=lambda b: b,
            base64_response_body=lambda b: base64.b64decode(bytes(b)),
            file_response_body=_get_file_contents,
            json_response_body=json.loads
        )

        # Set the testcase-specified python executable code
        create_compiler = lambda field_name: functools.partial(compile, filename="<%s field of %s>" % (field_name, t.name), mode="exec")
        t.on_request = field("on_request", create_compiler("on_request"))
        t.on_response = field("on_response", create_compiler("on_response"))

        return t

class Test(object):
    """Model that represents a test"""

    def __init__(self, name):
        """
        Creates a new test
        :arg string name: The name of the test
        """
        self.name = name
        self.testcases = []

    @classmethod
    def parse(cls, data):
        """
        Parses a test into a model
        :arg dict data: The data to parse into a model
        """
        
        field = functools.partial(_get_field, "test", data)
        test = cls(field("name", str, required=True))

        for testcase_data in field("testcases", list, required=True):
            test.testcases.append(Testcase.parse(testcase_data))

        return test



from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

version = "0.3.0"

from .models import ParseException, Testcase, Test, TestcaseResult
from .tabbing import tab, detab
from .worker import execute_testcase
from .yaml_parser import parse_yaml


from __future__ import absolute_import, division, print_function, with_statement, unicode_literals

import catnap
import yaml

def parse_yaml(f):
    """Parses a YAML-based test file"""
    return catnap.Test.parse(yaml.load(f))



from setuptools import setup

setup(
    name = "Catnap",
    version = "0.3.0",
    description = "A script for running integration tests against RESTful/HTTP-based interfaces",
    author = "Yusuf Simonson",
    url = "http://github.com/dailymuse/catnap",
    
    packages = [
        "catnap",
    ],
    
    scripts = ["scripts/catnap"],

    install_requires = [
        "PyYAML==3.10",
        "requests==1.2.3",
        "optfn==0.4.0"
    ]
)



import sys

if sys.version_info >= (3,):
    from io import StringIO
    import urllib.parse as urllib

    _oldbytes = bytes
    bytes = lambda b: _oldbytes(b, "ascii")
    str = str
else:
    try:
        from cStringIO import StringIO
    except:
        from StringIO import StringIO

    import urllib
    bytes = str
    str = unicode



import unittest
import catnap
from catnap.compat import *
import requests.auth

class TestModelTest(unittest.TestCase):
    def test_missing_name(self):
        self.assertRaises(catnap.ParseException, catnap.Test.parse, {"testcases": []})

    def test_missing_testcases(self):
        self.assertRaises(catnap.ParseException, catnap.Test.parse, {"name": "foo"})

    def test_valid(self):
        test = catnap.Test.parse({
            "name": "test",
            "testcases": [{
                "name": "minimal test",
                "url": "http://www.google.com",
            }, {
                "name": "complex test",
                "method": "POST",
                "url": "http://www.google.com",
                "query_params": { "hello": "world" },
                "headers": { "foo": "bar" },
                "auth": "basic user pass",
                "form_body": { "form_key": "form_value" },

                "code": 200,
                "response_url": "http://www.google.com?redirect",
                "response_headers": { "foo": "baz" },
                "base64_response_body": "aGVsbG8=",
                "on_request": "import sys; sys.stdout.write('request')",
                "on_response": "import sys; sys.stdout.write('response')",
            }]
        })

        self.assertEqual(test.name, "test")
        self.assertEqual(len(test.testcases), 2)

        t1 = test.testcases[0]

        self.assertEqual(t1.name, "minimal test")
        self.assertEqual(t1.method, "GET")
        self.assertEqual(t1.url, "http://www.google.com")
        self.assertEqual(t1.query_params, {})
        self.assertEqual(t1.headers, {})
        self.assertEqual(t1.auth, None)
        self.assertEqual(t1.body_type, None)
        self.assertEqual(t1.body, None)
        self.assertEqual(t1.code, None)
        self.assertEqual(t1.response_url, None)
        self.assertEqual(t1.response_headers, {})
        self.assertEqual(t1.response_body_type, None)
        self.assertEqual(t1.response_body, None)
        self.assertEqual(t1.on_request, None)
        self.assertEqual(t1.on_response, None)

        t2 = test.testcases[1]

        self.assertEqual(t2.name, "complex test")
        self.assertEqual(t2.method, "POST")
        self.assertEqual(t2.url, "http://www.google.com")
        self.assertEqual(t2.query_params, { "hello": "world" })
        self.assertEqual(t2.headers, { "foo": "bar" })
        self.assertEqual(t2.auth.username, "user")
        self.assertEqual(t2.auth.password, "pass")
        self.assertEqual(t2.body_type, "form_body")
        self.assertEqual(t2.body, "form_key=form_value")
        self.assertEqual(t2.code, 200)
        self.assertEqual(t2.response_url, "http://www.google.com?redirect")
        self.assertEqual(t2.response_headers, { "foo": "baz" })
        self.assertEqual(t2.response_body_type, "base64_response_body")
        self.assertEqual(t2.response_body, bytes("hello"))
        self.assertNotEqual(t2.on_request, None)
        self.assertNotEqual(t2.on_response, None)

class TestcaseModelTest(unittest.TestCase):
    def build_testcase(self, **kwargs):
        data = {
            "name": "foo",
            "url": "http://www.google.com",
        }

        data.update(kwargs)
        return catnap.Testcase.parse(data)

    def test_name(self):
        testcase = self.build_testcase()
        self.assertEqual(testcase.name, "foo")
        self.assertRaises(catnap.ParseException, catnap.Testcase.parse, {"url": "foo"})

    def test_method(self):
        testcase = self.build_testcase(method="get")
        self.assertEqual(testcase.method, "GET")
        self.assertRaises(catnap.ParseException, catnap.Testcase.parse, {"method": None})

    def test_url(self):
        testcase = self.build_testcase()
        self.assertEqual(testcase.url, "http://www.google.com")
        self.assertRaises(catnap.ParseException, catnap.Testcase.parse, {"name": "foo"})

    def test_query_params(self):
        testcase = self.build_testcase(query_params={"hello": "world"})
        self.assertEqual(testcase.query_params, {"hello": "world"})
        self.assertRaises(catnap.ParseException, self.build_testcase, query_params=None)

    def test_headers(self):
        testcase = self.build_testcase(headers={"hello": "world"})
        self.assertEqual(testcase.headers, {"hello": "world"})
        self.assertRaises(catnap.ParseException, self.build_testcase, headers=None)

    def test_basic_auth(self):
        testcase = self.build_testcase(auth="basic user pass")
        self.assertTrue(isinstance(testcase.auth, requests.auth.HTTPBasicAuth))
        self.assertEqual(testcase.auth.username, "user")
        self.assertEqual(testcase.auth.password, "pass")

    def test_digest_auth(self):
        testcase = self.build_testcase(auth="digest user pass")
        self.assertTrue(isinstance(testcase.auth, requests.auth.HTTPDigestAuth))
        self.assertEqual(testcase.auth.username, "user")
        self.assertEqual(testcase.auth.password, "pass")

    def test_bad_auth(self):
        self.assertRaises(catnap.ParseException, self.build_testcase, auth="unknown user pass")

    def test_body(self):
        testcase = self.build_testcase(body="body test")
        self.assertEqual(testcase.body, "body test")

    def test_form_body(self):
        testcase = self.build_testcase(form_body={"hello": "world", "foo": "bar"})
        self.assertTrue(testcase.body in ["hello=world&foo=bar", "foo=bar&hello=world"])
        self.assertRaises(catnap.ParseException, self.build_testcase, form_body="invalidbody")

    def test_base64_body(self):
        testcase = self.build_testcase(base64_body="aGVsbG8=")
        self.assertEqual(testcase.body, bytes("hello"))
        self.assertRaises(catnap.ParseException, self.build_testcase, base64_body="invalid base64 string")

    def test_file_body(self):
        testcase = self.build_testcase(file_body="./test/static/sample.txt")
        self.assertEqual(testcase.body, "hello from a sample text file")
        self.assertRaises(catnap.ParseException, self.build_testcase, file_body="./test/nonfile.txt")

    def test_invalid_multibody(self):
        self.assertRaises(catnap.ParseException, self.build_testcase, body="plaintext body", file_body="./test/nonfile.txt")

    def test_valid_code(self):
        testcase = self.build_testcase(code="400")
        self.assertEqual(testcase.code, 400)

    def test_invalid_code(self):
        self.assertRaises(catnap.ParseException, self.build_testcase, code=None)

    def test_response_body(self):
        testcase = self.build_testcase(response_body="body test")
        self.assertEqual(testcase.response_body_type, "response_body")
        self.assertEqual(testcase.response_body, "body test")

    def test_response_url(self):
        testcase = self.build_testcase(response_url="foobar")
        self.assertEqual(testcase.response_url, "foobar")
        self.assertRaises(catnap.ParseException, catnap.Testcase.parse, {"name": "foo", "response_url": None})

    def test_response_headers(self):
        testcase = self.build_testcase(response_headers={"hello": "world"})
        self.assertEqual(testcase.response_headers, {"hello": "world"})
        self.assertRaises(catnap.ParseException, self.build_testcase, response_headers=None)

    def test_base64_response_body(self):
        testcase = self.build_testcase(base64_response_body="aGVsbG8=")
        self.assertEqual(testcase.response_body_type, "base64_response_body")
        self.assertEqual(testcase.response_body, bytes("hello"))
        self.assertRaises(catnap.ParseException, self.build_testcase, base64_response_body="invalid base64 string")

    def test_file_response_body(self):
        testcase = self.build_testcase(file_response_body="./test/static/sample.txt")
        self.assertEqual(testcase.response_body_type, "file_response_body")
        self.assertEqual(testcase.response_body, "hello from a sample text file")
        self.assertRaises(catnap.ParseException, self.build_testcase, file_response_body="./test/nonfile.txt")

    def test_invalid_response_multibody(self):
        self.assertRaises(catnap.ParseException, self.build_testcase, response_body="plaintext body", file_response_body="./test/nonfile.txt")

    def test_valid_on_request(self):
        testcase = self.build_testcase(on_request="import sys; sys.stdout.write('hi')")
        self.assertNotEqual(testcase.on_request, None)

    def test_invalid_on_request(self):
        self.assertRaises(catnap.ParseException, self.build_testcase, on_request="!!!")

    def test_valid_on_response(self):
        testcase = self.build_testcase(on_response="import sys; sys.stdout.write('hi')")
        self.assertNotEqual(testcase.on_response, None)

    def test_invalid_on_response(self):
        self.assertRaises(catnap.ParseException, self.build_testcase, on_response="!!!")

if __name__ == "__main__":
    unittest.main()
